# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import logging
import json

logger = logging.getLogger('koppeltaal.connector')
requests_logger = logging.getLogger("requests.packages.urllib3")

critical = logger.critical
debug = logger.debug
error = logger.error
info = logger.info
warning = logger.warning


def set_log_level(level):
    logger.setLevel(level)
    requests_logger.setLevel(level)


def debug_json(message, **data):
    if 'json' in data:
        data['json'] = json.dumps(data['json'], indent=2, sort_keys=True)
    debug(message.format(**data))



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import lxml.etree
import six

from koppeltaal import fhir


unicode = six.text_type


TYPES = [
    (('extension', 'valueInteger'), 'int'),
    (('extension', 'valueDecimal'), 'float'),
    (('extension', 'valueBoolean'), 'boolean'),
    (('activity', 'prohibited'), 'boolean'),
    (('Patient', 'active'), 'boolean'),
]


def type_for(node_tag, child_tag):
    # Replace this by checking the definition.
    for key, value in TYPES:
        if (node_tag, child_tag) == key:
            return value
    return 'string'


def tag(node, namespace="atom"):
    tag = node.tag
    if namespace == "atom":
        namespace = '{http://www.w3.org/2005/Atom}'
    elif namespace == "fhir":
        namespace = '{http://hl7.org/fhir}'
    if tag.startswith(namespace):
        return tag[len(namespace):]
    return tag


def fhir2json(node, repeatable_field_names):
    node_tag = tag(node, "fhir")
    result = dict(node.attrib)
    for child in node.getchildren():
        child_tag = tag(child, "fhir")
        if child_tag == '{http://www.w3.org/1999/xhtml}div':
            child_tag = 'div'
            child_value = lxml.etree.tostring(child)
        elif len(child.getchildren()):
            child_value = fhir2json(child, repeatable_field_names)
            child_value.update(child.attrib)
        else:
            assert 'value' in child.attrib
            child_value_type = type_for(node_tag, child_tag)
            if child_value_type == 'int':
                child_value = int(child.attrib['value'])
            elif child_value_type == 'float':
                child_value = float(child.attrib['value'])
            elif child_value_type == 'boolean':
                child_value = child.attrib['value'] == 'true'
            else:
                child_value = unicode(child.attrib['value'])
        if child_tag in repeatable_field_names:
            result.setdefault(child_tag, []).append(child_value)
        else:
            assert child_tag not in result, \
                'Invalid fhir file: {} already present.'.format(child_tag)
            result[child_tag] = child_value
    return result


def atom2json(node, is_entry=False):
    result = {}
    for child in node.getchildren():
        child_tag = tag(child, "atom")
        if child_tag in {'id', 'updated', 'title'}:
            result[child_tag] = child.text
        if child_tag in {'category', 'link'}:
            child_value = result.setdefault(child_tag, [])
            child_value.append(dict(child.attrib))
        if child_tag == 'entry':
            assert not is_entry, 'Invalid atom file'
            child_value = result.setdefault(child_tag, [])
            child_value.append(atom2json(child, is_entry=True))
        if child_tag == 'content':
            assert is_entry, 'Invalid atom file'
            assert len(child.getchildren()), 'Invalid fhir file'
            fhir_node = child.getchildren()[0]
            fhir_type = tag(fhir_node, "fhir")
            field_names = fhir.REGISTRY.repeatable_field_names(fhir_type)
            child_value = {'resourceType': fhir_type}
            child_value.update(fhir2json(fhir_node, field_names))
            result[child_tag] = child_value
    return result


def xml2json(xml_file):
    feed = lxml.etree.parse(xml_file).getroot()
    assert tag(feed, "atom") == 'feed', 'Invalid atom file'

    bundle = {'resourceType': 'Bundle'}
    bundle.update(atom2json(feed))
    return bundle



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import koppeltaal.interfaces
import koppeltaal.definitions


class Code(dict):

    def __init__(self, system, items):
        super(Code, self).__init__(items)
        if not system.startswith('http:'):
            system = koppeltaal.interfaces.NAMESPACE + system
        self.system = system

    def pack_code(self, value):
        if value not in self:
            raise koppeltaal.interfaces.InvalidCode(self, value)
        return value

    def pack_coding(self, value):
        if value not in self:
            raise koppeltaal.interfaces.InvalidCode(self, value)
        display = self[value]
        coding = {"code": value, "system": self.system}
        if display is not None:
            coding["display"] = display
        return coding

    def unpack_code(self, code):
        if code not in self:
            raise koppeltaal.interfaces.InvalidCode(self, code)
        return code

    def unpack_coding(self, coding):
        value = coding.get("code")
        system = coding.get("system")
        if system == koppeltaal.definitions.NULL_SYSTEM \
                and value == koppeltaal.definitions.NULL_VALUE:
            return None
        if system != self.system:
            raise koppeltaal.interfaces.InvalidSystem(self, system)
        if value not in self:
            raise koppeltaal.interfaces.InvalidCode(self, value)
        return value


ACTIVITY_KIND = Code(
    'ActivityKind',
    {'ELearning': 'E-Learning',
     'Game': 'Game',
     'Meeting': 'Meeting',
     'MultipleActivityTemplate': 'Multiple activity template',
     'Questionnaire': 'Questionnaire',
     })

ACTIVITY_LAUNCH_TYPE = Code(
    'ActivityDefinitionLaunchType',
    {'Mobile': 'Mobile',
     'None': 'None',
     'Web': 'Web',
     })

ACTIVITY_PERFORMER = Code(
    'ActivityPerformer',
    {'Practitioner': 'Practitioner',
     'RelatedPerson': 'RelatedPerson',
     'Patient': 'Patient',
     })

ADDRESS_USE = Code(
    'http://hl7.org/fhir/address-use',
    {'home': 'home',
     'old': 'old',
     'temp': 'temp',
     'work': 'work',
     })

CAREPLAN_ACTIVITY_CATEGORY = Code(
    'http://hl7.org/fhir/care-plan-activity-category',
    {'diet': 'diet',
     'drug': 'drug',
     'encounter': 'encounter',
     'observation': 'observation',
     'other': 'other',
     'procedure': 'procedure',
     'supply': 'supply',
     })

CAREPLAN_PARTICIPANT_ROLE = Code(
    'CarePlanParticipantRole',
    {'Analyst': 'Analyst',
     'Caregiver': 'Caregiver',
     'Secretary': 'Secretary',
     'Supervisor': 'Supervisor',
     'Thirdparty': 'Thirdparty',
     'Requester': 'Requester',
     })

CAREPLAN_ACTIVITY_STATUS = Code(
    'CarePlanActivityStatus',
    {'Available': 'Available',
     'Cancelled': 'Cancelled',
     'Completed': 'Completed',
     'Expired': 'Expired',
     'InProgress': 'InProgress',
     'SkippedByUser': 'SkippedByUser',
     'Waiting': 'Waiting',
     })

CAREPLAN_GOAL_STATUS = Code(
    'http://hl7.org/fhir/care-plan-goal-status',
    {'achieved': 'achieved',
     'cancelled': 'cancelled',
     'in progress': 'in progress',
     'sustaining': 'sustaining',
     })

CAREPLAN_STATUS = Code(
    'http://hl7.org/fhir/care-plan-status',
    {'active': 'active',
     'completed': 'completed',
     'planned': 'planned',
     })

CONTACT_SYSTEM = Code(
    'http://hl7.org/fhir/contact-system',
    {'email': 'email',
     'fax': 'fax',
     'phone': 'phone',
     'url': 'url',
     })

CARE_TEAM_STATUS = Code(
    'CareTeamStatus',
    {'active': 'Active',
     'entered-in-error': 'Entered in error',
     'inactive': 'Inactive',
     'proposed': 'Proposed',
     'suspended': 'Suspended',
     })

CONTACT_ENTITY_TYPE = Code(
    'http://hl7.org/fhir/contactentity-type',
    {'ADMIN': 'ADMIN',
     'BILL': 'BILL',
     'HR': 'HR',
     'PATINF': 'PATINF',
     'PAYOR': 'PAYOR',
     'PRESS': 'PRESS',
     })

CONTACT_USE = Code(
    'http://hl7.org/fhir/contact-use',
    {'home': 'home',
     'mobile': 'mobile',
     'old': 'old',
     'temp': 'temp',
     'work': 'work',
     })

DEVICE_KIND = Code(
    'DeviceKind',
    {'Application': 'Application'})

GENDER = Code(
    'http://hl7.org/fhir/v3/AdministrativeGender',
    {'F': 'Female',
     'M': 'Male',
     'UN': 'Undifferentiated',
     })

IDENTIFIER_USE = Code(
    'http://hl7.org/fhir/identifier-use',
    {'official': 'official',
     'secondary': 'secondary',
     'temp': 'temp',
     'usual': 'usual',
     })

MESSAGE_HEADER_RESPONSE_CODE = Code(
    'http://hl7.org/fhir/response-code',
    {'fatal-error': 'fatal-error',
     'ok': 'ok',
     'transient-error': 'transient-error',
     })

MESSAGE_HEADER_EVENTS = Code(
    'MessageEvents',
    {'CreateOrUpdateActivityDefinition': 'CreateOrUpdateActivityDefinition',
     'CreateOrUpdateCarePlan': 'CreateOrUpdateCarePlan',
     'CreateOrUpdateCarePlanActivityResult':
        'CreateOrUpdateCarePlanActivityResult',
     'CreateOrUpdatePatient': 'CreateOrUpdatePatient',
     'CreateOrUpdatePractitioner': 'CreateOrUpdatePractitioner',
     'CreateOrUpdateRelatedPerson': 'CreateOrUpdateRelatedPerson',
     'CreateOrUpdateUserMessage': 'CreateOrUpdateUserMessage',
     'UpdateCarePlanActivityStatus': 'UpdateCarePlanActivityStatus',
     })

MESSAGE_KIND = Code(
    'UserMessageKind',
    {'Advice': 'Advice',
     'Alert': 'Alert',
     'Answer': 'Answer',
     'Message': 'Message',
     'Notification': 'Notification',
     'Question': 'Question',
     'Request': 'Request',
     })

NAME_USE = Code(
    'http://hl7.org/fhir/name-use',
    {'anonymous': 'anonymous',
     'maiden': 'maiden',
     'nickname': 'nickname',
     'official': 'official',
     'old': 'old',
     'temp': 'temp',
     'usual': 'usual',
     })

ORGANIZATION_TYPE = Code(
    'http://hl7.org/fhir/organization-type',
    {'dept': 'dept',
     'edu': 'edu',
     'fed': 'fed',
     'icu': 'icu',
     'ins': 'ins',
     'pharm': 'pharm',
     'reli': 'reli',
     'team': 'team',
     })

RELATED_PERSON_RELATIONSHIP_TYPE = Code(
    'http://hl7.org/fhir/vs/relatedperson-relationshiptype',
    {'emergency': 'Emergency',
     'family': 'Family',
     'guardian': 'Guardian',
     'friend': 'Friend',
     'partner': 'Partner',
     'work': 'Work',
     'caregiver': 'Caregiver',
     'agent': 'Agent',
     'guarantor': 'Guarantor',
     'owner': 'Owner of animal',
     'parent': 'Parent',
     'FAMMEMB': 'family member',
     'CHILD': 'child',
     'CHLDADOPT': 'adopted child',
     'DAUADOPT': 'adopted daughter',
     'SONADOPT': 'adopted son',
     'CHLDFOST': 'foster child',
     'DAUFOST': 'foster daughter',
     'SONFOST': 'foster son',
     'DAUC': 'daughter',
     'DAU': 'natural daughter',
     'STPDAU': 'stepdaughter',
     'NCHILD': 'natural child',
     'SON': 'natural son',
     'SONC': 'son',
     'STPSON': 'stepson',
     'STPCHLD': 'step child',
     'EXT': 'extended family member',
     'AUNT': 'aunt',
     'MAUNT': 'maternal aunt',
     'PAUNT': 'paternal aunt',
     'COUSN': 'cousin',
     'MCOUSN': 'maternal cousin',
     'PCOUSN': 'paternal cousin',
     'GGRPRN': 'great grandparent',
     'GGRFTH': 'great grandfather',
     'MGGRFTH': 'maternal great-grandfather',
     'PGGRFTH': 'paternal great-grandfather',
     'GGRMTH': 'great grandmother',
     'MGGRMTH': 'maternal great-grandmother',
     'PGGRMTH': 'paternal great-grandmother',
     'MGGRPRN': 'maternal great-grandparent',
     'PGGRPRN': 'paternal great-grandparent',
     'GRNDCHILD': 'grandchild',
     'GRNDDAU': 'granddaughter',
     'GRNDSON': 'grandson',
     'GRPRN': 'grandparent',
     'GRFTH': 'grandfather',
     'MGRFTH': 'maternal grandfather',
     'PGRFTH': 'paternal grandfather',
     'GRMTH': 'grandmother',
     'MGRMTH': 'maternal grandmother',
     'PGRMTH': 'paternal grandmother',
     'MGRPRN': 'maternal grandparent',
     'PGRPRN': 'paternal grandparent',
     'INLAW': 'inlaw',
     'CHLDINLAW': 'child-in-law',
     'DAUINLAW': 'daughter in-law',
     'SONINLAW': 'son in-law',
     'PRNINLAW': 'parent in-law',
     'FTHINLAW': 'father-in-law',
     'MTHINLAW': 'mother-in-law',
     'SIBINLAW': 'sibling in-law',
     'BROINLAW': 'brother-in-law',
     'SISINLAW': 'sister-in-law',
     'NIENEPH': 'niece/nephew',
     'NEPHEW': 'nephew',
     'NIECE': 'niece',
     'UNCLE': 'uncle',
     'MUNCLE': 'maternal uncle',
     'PUNCLE': 'paternal uncle',
     'PRN': 'parent',
     'ADOPTP': 'adoptive parent',
     'ADOPTF': 'adoptive father',
     'ADOPTM': 'adoptive mother',
     'FTH': 'father',
     'FTHFOST': 'foster father',
     'NFTH': 'natural father',
     'NFTHF': 'natural father of fetus',
     'STPFTH': 'stepfather',
     'MTH': 'mother',
     'GESTM': 'gestational mother',
     'MTHFOST': 'foster mother',
     'NMTH': 'natural mother',
     'NMTHF': 'natural mother of fetus',
     'STPMTH': 'stepmother',
     'NPRN': 'natural parent',
     'PRNFOST': 'foster parent',
     'STPPRN': 'step parent',
     'SIB': 'sibling',
     'BRO': 'brother',
     'HBRO': 'half-brother',
     'NBRO': 'natural brother',
     'TWINBRO': 'twin brother',
     'FTWINBRO': 'fraternal twin brother',
     'ITWINBRO': 'identical twin brother',
     'STPBRO': 'stepbrother',
     'HSIB': 'half-sibling',
     'HSIS': 'half-sister',
     'NSIB': 'natural sibling',
     'NSIS': 'natural sister',
     'TWINSIS': 'twin sister',
     'FTWINSIS': 'fraternal twin sister',
     'ITWINSIS': 'identical twin sister',
     'TWIN': 'twin',
     'FTWIN': 'fraternal twin',
     'ITWIN': 'identical twin',
     'SIS': 'sister',
     'STPSIS': 'stepsister',
     'STPSIB': 'step sibling',
     'SIGOTHR': 'significant other',
     'DOMPART': 'domestic partner',
     'SPS': 'spouse',
     'HUSB': 'husband',
     'WIFE': 'wife',
     'FRND': 'unrelated friend',
     'NBOR': 'neighbor',
     'ONESELF': 'self',
     'ROOM': 'Roommate'
     })

# BlackBoxState is not valid, but the javascript connector generate
# some of those.
OTHER_RESOURCE_USAGE = Code(
    'OtherResourceUsage',
    {'ActivityDefinition': 'ActivityDefinition',
     'BlackBoxState': 'BlackBoxState',
     'CarePlanActivityStatus': 'CarePlanActivityStatus',
     'CareTeam': 'CareTeam',
     'StorageItem': 'StorageItem',
     'UserMessage': 'UserMessage',
     })


PROCESSING_STATUS = Code(
    'ProcessingStatus',
    {'Claimed': 'Claimed',
     'Failed': 'Failed',
     'MaximumRetriesExceeded': 'MaximumRetriesExceeded',
     'New': 'New',
     'ReplacedByNewVersion': 'ReplacedByNewVersion',
     'Success': 'Success',
     })


ISSUE_TYPE = Code(
    'http://hl7.org/fhir/issue-type',
    {'invalid': None,
     'structure': None,
     'required': None,
     'value': None,
     'invariant': None,
     'security': None,
     'login': None,
     'unknown': None,
     'expired': None,
     'forbidden': None,
     'processing': None,
     'not-supported': None,
     'duplicate': None,
     'not-found': None,
     'too-long': None,
     'code-unknown': None,
     'extension': None,
     'too-costly': None,
     'business-rule': None,
     'conflict': None,
     'transient': None,
     'lock-error': None,
     'no-store': None,
     'exception': None,
     'timeout': None,
     'throttled': None,
     })


ISSUE_SEVERITY = Code(
    'http://hl7.org/fhir/issue-severity',
    {'fatal': None,
     'error': None,
     'warning': None,
     'information': None,
     })



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import datetime
import uuid
import pytest
import selenium.webdriver
import six
import koppeltaal.connector
import koppeltaal.models
import koppeltaal.testing
import koppeltaal.utils


unicode = six.text_type


def pytest_addoption(parser):
    '''Add server identifier to be passed in. Looks for corresponding part in
    ~/.koppeltaal.cfg
    '''
    parser.addoption(
        '--server',
        help=("""\
Koppeltaal server identifier. URL and credentials should be
defined in the part identified by that name in ~/.koppeltaal.cfg.

For example:

[edge]
url = https://edgekoppeltaal.vhscloud.nl
username = [username]
password = [password]
domain = PythonAdapterTesting
oauth_secret = [oauth password]

"""),
        default="edge")


@pytest.fixture(scope='session')
def connector(request):
    server = request.config.option.server
    if not server:
        raise ValueError("No server name defined.")
    credentials = koppeltaal.utils.get_credentials_from_file(server)
    integration = koppeltaal.connector.Integration(
        name='Python connector tests')
    return koppeltaal.connector.Connector(credentials, integration)


@pytest.fixture
def transport(monkeypatch, connector):
    transport = koppeltaal.testing.MockTransport('koppeltaal.tests')
    monkeypatch.setattr(connector, 'transport', transport)
    monkeypatch.setattr(connector.integration, 'model_id', lambda m: u'1')
    monkeypatch.setattr(koppeltaal.utils, 'messageid', lambda: u'1234-5678')
    return transport


@pytest.fixture
def patient(request, connector):
    patient = koppeltaal.models.Patient(
        name=[
            koppeltaal.models.Name(
                family=[u"Doe"],
                given=[u"John"],
                text=u'John Doe')],
        age=42,
        gender="M",
        active=True)
    return patient


@pytest.fixture
def practitioner():
    return koppeltaal.models.Practitioner(
        name=koppeltaal.models.Name(
            given=[u'John'],
            family=[u'Q.', u'Practitioner'],
            text=u'John Q Practitioner'))


@pytest.fixture
def related_person(patient):
    related_person = koppeltaal.models.RelatedPerson(
        patient=patient,
        relationship='PRN',
        name=koppeltaal.models.Name(
                family=[u"Doe"],
                given=[u"John"],
                text=u'John Doe'),
        gender="M",
        address=koppeltaal.models.Address(city='Rotterdam'))
    return related_person


@pytest.fixture
def activity_definition(connector):
    # Highly depending on what's activated on the server.
    definition = connector.activity('KTSTESTGAME')
    assert definition is not None, 'Test activity not found.'
    return definition


@pytest.fixture
def careplan(patient, practitioner, activity_definition):
    participants = [koppeltaal.models.Participant(
        member=practitioner,
        role='Caregiver')]
    return koppeltaal.models.CarePlan(
        activities=[koppeltaal.models.Activity(
            identifier=unicode(uuid.uuid4()),
            definition=activity_definition.identifier,
            kind=activity_definition.kind,
            participants=participants,
            planned=datetime.datetime.now(),
            status='Available')],
        patient=patient,
        participants=participants,
        status='active')


@pytest.fixture
def careplan_from_fixture(request, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/Other/_search?code=ActivityDefinition',
        respond_with='fixtures/activities_game.json')
    return request.getfixturevalue('careplan')


@pytest.fixture
def careplan_response(connector, careplan):
    return connector.send('CreateOrUpdateCarePlan', careplan, careplan.patient)


@pytest.fixture(scope='session')
def driver(request):
    driver = selenium.webdriver.Firefox()
    request.addfinalizer(driver.quit)
    return driver


@pytest.fixture
def browser(driver, request):
    request.addfinalizer(driver.delete_all_cookies)
    return driver



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import requests
import six

from koppeltaal import (interfaces, logger)
from six.moves.urllib.parse import urlparse, urlunparse


unicode = six.text_type


class Response(object):

    def __init__(self, json=None, location=None):
        self.json = json
        self.location = location


class Transport(object):

    def __init__(self, server, username, password):
        parts = urlparse(server)

        self.server = server
        self.scheme = parts.scheme
        self.netloc = parts.netloc
        self.username = username
        self.password = password

        self.session = requests.Session()

    def absolute_url(self, url):
        # Make sure we talk to the proper server by updating the URL.
        parts = list(map(unicode, urlparse(url)[2:]))
        return urlunparse([unicode(self.scheme), unicode(self.netloc)] + parts)

    def _read_http_response(self, http_response):
        if not http_response.headers['content-type'].startswith(
                'application/json'):
            raise interfaces.ConnectionError(http_response)
        response = Response(
            json=http_response.json() if http_response.text else None,
            location=http_response.headers.get('content-location'))
        if 400 <= http_response.status_code < 600:
            raise interfaces.ResponseError(response)
        return response

    def query(self, url, params=None, username=None, password=None):
        """Query a url.
        """
        try:
            http_response = self.session.get(
                self.absolute_url(url),
                params=params,
                auth=(username or self.username, password or self.password),
                headers={'Accept': 'application/json'},
                timeout=interfaces.TIMEOUT,
                allow_redirects=False)
        except requests.RequestException as error:
            raise interfaces.ConnectionError(error)
        return self._read_http_response(http_response)

    def query_redirect(self, url, params=None):
        """Query a url for a redirect.
        """
        try:
            http_response = self.session.get(
                self.absolute_url(url),
                params=params,
                auth=(self.username, self.password),
                timeout=interfaces.TIMEOUT,
                allow_redirects=False)
        except requests.RequestException as error:
            raise interfaces.ConnectionError(error)
        if not http_response.is_redirect:
            raise interfaces.ConnectionError(http_response)
        return Response(location=http_response.headers.get('location'))

    def create(self, url, data):
        """Create a new resource at the given url with JSON data.
        """
        try:
            http_response = self.session.post(
                self.absolute_url(url),
                auth=(self.username, self.password),
                json=data,
                headers={'Accept': 'application/json'},
                timeout=interfaces.TIMEOUT,
                allow_redirects=False)
        except requests.RequestException as error:
            raise interfaces.ConnectionError(error)
        return self._read_http_response(http_response)

    def update(self, url, data):
        """Update an existing resource at the given url with JSON data.
        """
        try:
            http_response = self.session.put(
                self.absolute_url(url),
                auth=(self.username, self.password),
                json=data,
                headers={'Accept': 'application/json'},
                timeout=interfaces.TIMEOUT,
                allow_redirects=False)
        except requests.RequestException as error:
            raise interfaces.ConnectionError(error)
        return self._read_http_response(http_response)

    def close(self):
        self.session.close()



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import datetime
import dateutil.parser
import six
import zope.interface

from koppeltaal import (
    fhir,
    compat,
    definitions,
    interfaces,
    models,
    utils)


basestring = six.string_types
unicode = six.text_type


@zope.interface.implementer(interfaces.IBrokenFHIRResource)
class BrokenResource(object):
    fhir_link = None
    error = None
    payload = None

    def __init__(self, error, payload):
        self.error = error
        self.payload = payload

    def __str__(self):
        return u"Broken resource '{}': {}".format(self.fhir_link, self.error)


class Extension(object):

    def __init__(self, packer, content=None):
        self._packer = packer
        self._index = {}
        if content and 'extension' in content:
            for extension in content['extension']:
                url = extension['url']
                self._index.setdefault(url, []).append(extension)
        compat.extensions(self._index)

    def _unpack_item(self, field, extension):
        if field.field_type == 'boolean':
            value = extension.get('valueBoolean')
            if not isinstance(value, bool):
                raise interfaces.InvalidValue(field, extension)
            return value

        if field.field_type == 'codeable':
            # Note how 'codeable' is actually an extension data type but we
            # treat it specially here.
            value = extension.get('valueCodeableConcept')
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, extension)
            if 'coding' not in value:
                raise interfaces.InvalidValue(field, extension)
            if not isinstance(value['coding'], list):
                raise interfaces.InvalidValue(field, extension)
            if len(value['coding']) != 1:
                raise interfaces.InvalidValue(field, extension)
            return field.binding.unpack_coding(value['coding'][0])

        if field.field_type == 'code':
            value = extension.get('valueCode')
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, extension)
            return field.binding.unpack_code(value)

        if field.field_type == 'coding':
            # Note how 'coding' is actually an extension data type but we
            # treat it specially here.
            value = extension.get('valueCoding')
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, extension)
            return field.binding.unpack_coding(value)

        if field.field_type == 'date':
            value = extension.get('valueDate')
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, extension)
            try:
                return dateutil.parser.parse(value).date()
            except ValueError:
                raise interfaces.InvalidValue(field, extension)

        if field.field_type == 'datetime':
            value = extension.get('valueDateTime')
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, extension)
            try:
                return dateutil.parser.parse(value)
            except ValueError:
                raise interfaces.InvalidValue(field, extension)

        if field.field_type == 'instant':
            value = extension.get('valueInstant')
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, extension)
            try:
                return dateutil.parser.parse(value)
            except ValueError:
                raise interfaces.InvalidValue(field, extension)

        if field.field_type == 'integer':
            value = extension.get('valueInteger')
            if not isinstance(value, int):
                raise interfaces.InvalidValue(field, extension)
            return value

        if field.field_type == 'object':
            key = field.binding.queryTaggedValue('extension data type')
            if key is None:
                value = extension.get('extension')
                if not isinstance(value, list):
                    raise interfaces.InvalidValue(field, extension)
                return self._packer.unpack(extension, field.binding)

            value = extension.get(key)
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, extension)
            return self._packer.unpack(value, field.binding)

        if field.field_type in {'reference', 'versioned reference'}:
            value = extension.get('valueResource')
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, extension)
            return self._packer.unpack_reference(value)

        if field.field_type == 'string':
            value = extension.get('valueString')
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, extension)
            return value

        raise NotImplementedError()

    def unpack(self, field):
        if field.url not in self._index:
            if field.optional:
                if field.multiple:
                    return []
                return field.default
            raise interfaces.RequiredMissing(field)
        extensions = self._index[field.url]
        if field.multiple:
            values = []
            for extension in extensions:
                values.append(self._unpack_item(field, extension))
            return values

        if len(extensions) != 1:
            raise interfaces.InvalidValue(field)
        return self._unpack_item(field, extensions[0])

    @property
    def payload(self):
        all_extensions = []
        for extensions in self._index.values():
            all_extensions.extend(extensions)
        if all_extensions:
            return {"extension": all_extensions}
        return {}

    def _pack_item(self, field, value):
        if field.field_type == 'boolean':
            if not isinstance(value, bool):
                raise interfaces.InvalidValue(field, value)
            return {'valueBoolean': value}

        if field.field_type == 'code':
            if not isinstance(value, basestring):
                raise interfaces.InvalidValue(field, value)
            return {'valueCode': field.binding.pack_code(value)}

        if field.field_type == 'codeable':
            # Note how 'codeable' is actually an extension data type but we
            # treat it specially here.
            if not isinstance(value, basestring):
                raise interfaces.InvalidValue(field, value)
            return {"valueCodeableConcept":
                    {"coding": [field.binding.pack_coding(value)]}}

        if field.field_type == 'coding':
            # Note how 'coding' is actually an extension data type but we
            # treat it specially here.
            if not isinstance(value, basestring):
                raise interfaces.InvalidValue(field, value)
            return {'valueCoding': field.binding.pack_coding(value)}

        if field.field_type == 'date':
            if not isinstance(value, datetime.date):
                raise interfaces.InvalidValue(field, value)
            return {'valueDate': value.isoformat()}

        if field.field_type == 'datetime':
            if not isinstance(value, datetime.datetime):
                raise interfaces.InvalidValue(field, value)
            return {'valueDateTime': value.isoformat()}

        if field.field_type == 'instant':
            if not isinstance(value, datetime.datetime):
                raise interfaces.InvalidValue(field, value)
            if value.tzinfo is None:
                # We need a timezone! We assume timezone-naive datetimes
                # represent UTC times. So we add the UTC tzinfo here.
                value = value.replace(tzinfo=utils.utc, microsecond=0)
            return {'valueInstant': value.isoformat()}

        if field.field_type == 'integer':
            if not isinstance(value, int):
                raise interfaces.InvalidValue(field, value)
            return {'valueInteger': value}

        if field.field_type == 'object':
            if not isinstance(value, object):
                raise interfaces.InvalidValue(field, value)
            key = field.binding.queryTaggedValue('extension data type')
            if key is None:
                return self._packer.pack(value, field.binding)
            return {key: self._packer.pack(value, field.binding)}

        if field.field_type == 'reference':
            if not isinstance(value, object):
                raise interfaces.InvalidValue(field, value)
            return {'valueResource': self._packer.pack_reference(value)}

        if field.field_type == 'versioned reference':
            if not isinstance(value, object):
                raise interfaces.InvalidValue(field, value)
            ref = self._packer.pack_reference(value, versioned=True)
            return {'valueResource': ref}

        if field.field_type == 'string':
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, value)
            return {'valueString': value}

        raise NotImplementedError()

    def pack(self, field, value):
        if field.is_empty(value):
            if not field.optional:
                raise interfaces.InvalidValue(field, value)
            return
        if field.multiple:
            if not isinstance(value, list):
                raise interfaces.InvalidValue(field, value)
        else:
            value = [value]
        for single_value in value:
            extension = {"url": field.url}
            extension.update(self._pack_item(field, single_value))
            self._index.setdefault(field.url, []).append(extension)


class Native(object):

    def __init__(self, packer, content=None):
        self._packer = packer
        self._content = content or {}

    @property
    def payload(self):
        return self._content.copy()

    def _unpack_item(self, field, value):
        if field.field_type == 'boolean':
            if not isinstance(value, bool):
                raise interfaces.InvalidValue(field, value)
            return value

        if field.field_type == 'codeable':
            # Note how 'codeable' is actually an extension data type but we
            # treat it specially here.
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, value)
            if 'coding' not in value:
                raise interfaces.InvalidValue(field, value)
            if not isinstance(value['coding'], list):
                raise interfaces.InvalidValue(field, value)
            if len(value['coding']) != 1:
                raise interfaces.InvalidValue(field, value)
            return field.binding.unpack_coding(value['coding'][0])

        if field.field_type == 'code':
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, value)
            return field.binding.unpack_code(value)

        if field.field_type == 'coding':
            # Note how 'coding' is actually an extension data type but we
            # treat it specially here.
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, value)
            return field.binding.unpack_coding(value)

        if field.field_type == 'date':
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, value)
            try:
                return dateutil.parser.parse(value).date()
            except ValueError:
                raise interfaces.InvalidValue(field, value)

        if field.field_type == 'datetime':
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, value)
            try:
                return dateutil.parser.parse(value)
            except ValueError:
                raise interfaces.InvalidValue(field, value)

        if field.field_type == 'instant':
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, value)
            try:
                return dateutil.parser.parse(value)
            except ValueError:
                raise interfaces.InvalidValue(field, value)

        if field.field_type == 'integer':
            if not isinstance(value, int):
                raise interfaces.InvalidValue(field, value)
            return value

        if field.field_type == 'object':
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, value)
            return self._packer.unpack(value, field.binding)

        if field.field_type in {'reference', 'versioned reference'}:
            if not isinstance(value, dict):
                raise interfaces.InvalidValue(field, value)
            return self._packer.unpack_reference(value)

        if field.field_type == 'string':
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, value)
            return value

        raise NotImplementedError()

    def unpack(self, field):
        if field.name not in self._content:
            if field.optional:
                if field.multiple:
                    return []
                return field.default
            raise interfaces.RequiredMissing(field)

        value = self._content[field.name]
        if field.multiple:
            # If the field is multiple there is a list of item. We
            # only support the first one at the moment.
            if not isinstance(value, list):
                raise interfaces.InvalidValue(field, value)
            if not len(value):
                raise interfaces.RequiredMissing(field)
            if field.multiple:
                return [self._unpack_item(field, v) for v in value]
            value = value[0]
        return self._unpack_item(field, value)

    def _pack_item(self, field, value):
        if field.field_type == 'boolean':
            if not isinstance(value, bool):
                raise interfaces.InvalidValue(field, value)
            return value

        if field.field_type == 'codeable':
            # Note how 'codeable' is actually an extension data type but we
            # treat it specially here.
            if not isinstance(value, basestring):
                raise interfaces.InvalidValue(field, value)
            return {"coding": [field.binding.pack_coding(value)]}

        if field.field_type == 'code':
            if not isinstance(value, basestring):
                raise interfaces.InvalidValue(field, value)
            return field.binding.pack_code(value)

        if field.field_type == 'coding':
            # Note how 'coding' is actually an extension data type but we
            # treat it specially here.
            if not isinstance(value, basestring):
                raise interfaces.InvalidValue(field, value)
            return field.binding.pack_coding(value)

        if field.field_type == 'date':
            if not isinstance(value, datetime.date):
                raise interfaces.InvalidValue(field, value)
            return value.isoformat()

        if field.field_type == 'datetime':
            if not isinstance(value, datetime.datetime):
                raise interfaces.InvalidValue(field, value)
            return value.isoformat()

        if field.field_type == 'instant':
            if not isinstance(value, datetime.datetime):
                raise interfaces.InvalidValue(field, value)
            if value.tzinfo is None:
                value = value.replace(tzinfo=utils.utc, microsecond=0)
            return value.isoformat()

        if field.field_type == 'integer':
            if not isinstance(value, int):
                raise interfaces.InvalidValue(field, value)
            return value

        if field.field_type == 'object':
            if not isinstance(value, object):
                raise interfaces.InvalidValue(field, value)
            return self._packer.pack(value, field.binding)

        if field.field_type == 'reference':
            if not isinstance(value, object):
                raise interfaces.InvalidValue(field, value)
            return self._packer.pack_reference(value)

        if field.field_type == 'versioned reference':
            if not isinstance(value, object):
                raise interfaces.InvalidValue(field, value)
            return self._packer.pack_reference(value, versioned=True)

        if field.field_type == 'string':
            if not isinstance(value, unicode):
                raise interfaces.InvalidValue(field, value)
            return value

        raise NotImplementedError()

    def pack(self, field, value):
        if field.is_empty(value):
            if not field.optional:
                raise interfaces.InvalidValue(field, value)
            return
        if field.multiple:
            if not isinstance(value, list):
                raise interfaces.InvalidValue(field, value)
            item = [self._pack_item(field, v) for v in value]
        else:
            assert field.multiple is False
            item = self._pack_item(field, value)
        self._content[field.name] = item


class Packer(object):

    def __init__(self, resource, fhir_link):
        self.resource = resource
        self.fhir_link = fhir_link
        self._idref = 0

    def idref(self):
        self._idref += 1
        return 'ref{0:03}'.format(self._idref)

    def unpack(self, payload, definition, allow_broken=False):
        factory = fhir.REGISTRY.model_for_definition(definition)
        if factory is None:
            return None

        try:
            extension = Extension(self, payload)
            native = Native(self, payload)
            data = {}
            for name, field in definition.namesAndDescriptions():
                if not isinstance(field, definitions.Field):
                    continue
                if field.extension is None:
                    data[name] = native.unpack(field)
                else:
                    data[name] = extension.unpack(field)
            return factory(**data)
        except interfaces.InvalidValue as error:
            if allow_broken:
                return BrokenResource(error, payload)
            raise

    def pack(self, model, definition):
        extension = Extension(self)
        native = Native(self)

        if not definition.providedBy(model):
            raise interfaces.InvalidResource(definition, model)

        for name, field in definition.namesAndDescriptions():
            if not isinstance(field, definitions.Field):
                continue
            value = getattr(model, name, field.default)
            if field.extension is None:
                native.pack(field, value)
            else:
                extension.pack(field, value)
        # We do not have to add an idref because we do not refer back to
        # any object. However due to a bug the Javascript connector
        # requires it in some cases.
        payload = {'id': self.idref()}
        payload.update(extension.payload)
        payload.update(native.payload)
        return payload

    def unpack_reference(self, value):
        reference = self.resource.find(value)
        if reference:
            return reference.unpack()
        if not ('reference' in value or 'display' in value):
            raise interfaces.InvalidReference(value)
        return models.ReferredResource(
            fhir_link=value.get('reference'),
            display=value.get('display'))

    def pack_reference(self, value, versioned=False):
        if interfaces.IReferredFHIRResource.providedBy(value):
            reference = {}
            if value.fhir_link:
                reference['reference'] = value.fhir_link
            if value.display:
                reference['display'] = value.display
            if not reference:
                raise interfaces.InvalidReference(value)
            return reference
        entry = self.resource.add_model(value)
        ref = entry.fhir_link
        if not versioned:
            ref = utils.strip_history_from_link(ref)
        return {'reference': ref}



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import zope.interface.verify
import koppeltaal.definitions
import koppeltaal.interfaces
import koppeltaal.models
import koppeltaal.testing


def test_activities_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/Other/_search?code=ActivityDefinition',
        respond_with='fixtures/activities_game.json')

    activities = list(connector.activities())
    assert len(activities) == 2

    activity1, activity2 = activities

    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.ActivityDefinition, activity1)
    assert activity1.identifier == 'KTSTESTGAME'
    assert activity1.name == 'Test game'
    assert activity1.kind == 'Game'
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource, activity1.application)

    assert activity1.description == 'Testtest'
    assert activity1.fhir_link == (
        'https://edgekoppeltaal.vhscloud.nl/FHIR/Koppeltaal/Other/'
        'ActivityDefinition:3/_history/2016-07-04T07:49:01:742.1933')
    assert activity1.is_active is False
    assert activity1.is_archived is False
    assert activity1.is_domain_specific is False
    assert activity1.launch_type == 'Web'
    assert activity1.performer == 'Patient'
    assert activity1.subactivities == []

    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.ActivityDefinition, activity2)
    assert activity2.identifier == 'RANJKA'
    assert activity2.name == 'Ranj Kick ASS Game'
    assert activity2.kind == 'Game'

    assert activity2.description is None
    assert activity2.fhir_link == (
        'https://edgekoppeltaal.vhscloud.nl/FHIR/Koppeltaal/Other/'
        'ActivityDefinition:1433/_history/2016-07-04T09:04:24:679.3465')
    assert activity2.is_active is True
    assert activity2.is_archived is False
    assert activity2.is_domain_specific is True
    assert activity2.launch_type == u'Web'
    assert activity2.performer == u'Patient'
    assert len(activity2.subactivities) == 12
    for subactivity in activity2.subactivities:
        assert zope.interface.verify.verifyObject(
            koppeltaal.definitions.SubActivityDefinition, subactivity)

    assert subactivity.active is False
    assert subactivity.description == (
        u"Het is vrijdagmiddag en de werkweek is bijna voorbij. Maar niet "
        "voor jou, want jij moet voor vijf uur handgeschreven notulen "
        "hebben uitgewerkt. Je collega's treffen ondertussen de "
        "voorbereidingen voor het vieren van de verjaardag van hun chef.")
    assert subactivity.identifier == u'scenario_12'
    assert subactivity.name == u'Verjaardag op het werk'


def test_activity_from_fixture(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/Other/_search?code=ActivityDefinition',
        respond_with='fixtures/activities_game.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/Other/_search?code=ActivityDefinition',
        respond_with='fixtures/activities_game.json')

    activity = connector.activity('FOO')
    assert activity is None

    activity = connector.activity('KTSTESTGAME')
    assert activity is not None
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.ActivityDefinition, activity)
    assert activity.identifier == 'KTSTESTGAME'
    assert activity.name == 'Test game'
    assert activity.kind == 'Game'


def test_createorupdate_activitydefinition(connector, transport):
    transport.expect(
        'POST',
        '/FHIR/Koppeltaal/Mailbox',
        respond_with='fixtures/bundle_post_activitydefinition_ok.json')

    definition = koppeltaal.models.ActivityDefinition(
        application=koppeltaal.models.ReferredResource(display='Foobar'),
        description=u'A First Activity Definition',
        identifier=u'foobar',
        kind='ELearning',
        name=u'AD-1',
        performer='Patient',
        subactivities=[])

    response_data = connector.send(
        'CreateOrUpdateActivityDefinition', definition)
    assert len(response_data) == 1
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource, response_data[0])
    assert response_data[0].fhir_link == (
        'https://example.com/fhir/Koppeltaal/ActivityDefinition/1/'
        '_history/1970-01-01T01:01:01:01.1')

    response = transport.called.get('/FHIR/Koppeltaal/Mailbox')
    assert response is not None


def test_create_activitydefinition(connector, transport):
    transport.expect(
        'POST',
        koppeltaal.interfaces.OTHER_URL,
        redirect_to=(
            'https://example.com/fhir/Koppeltaal/ActivityDefinition/1/'
            '_history/1970-01-01T01:01:01:01.1'))

    definition = koppeltaal.models.ActivityDefinition(
        application=koppeltaal.models.ReferredResource(display='Foobar'),
        description=u'A First Activity Definition',
        identifier=u'foobar',
        kind='ELearning',
        name=u'AD-1',
        performer='Patient',
        subactivities=[])

    definition = connector.send_activity(definition)
    assert definition.fhir_link == (
        'https://example.com/fhir/Koppeltaal/ActivityDefinition/1/'
        '_history/1970-01-01T01:01:01:01.1')


def test_update_activitydefinition(connector, transport):
    transport.expect(
        'PUT',
        '/fhir/Koppeltaal/ActivityDefinition/1/'
        '_history/1970-01-01T01:01:01:01.1',
        redirect_to=(
            'https://example.com/fhir/Koppeltaal/ActivityDefinition/1/'
            '_history/1970-02-02T02:02:02:02.2'))

    definition = koppeltaal.models.ActivityDefinition(
        application=koppeltaal.models.ReferredResource(display='Foobar'),
        description=u'A First Activity Definition',
        identifier=u'foobar',
        kind='ELearning',
        name=u'AD-1',
        performer='Patient',
        subactivities=[])
    definition.fhir_link = (
        'https://example.com/fhir/Koppeltaal/ActivityDefinition/1/'
        '_history/1970-01-01T01:01:01:01.1')

    definition = connector.send_activity(definition)
    assert definition.fhir_link == (
        'https://example.com/fhir/Koppeltaal/ActivityDefinition/1/'
        '_history/1970-02-02T02:02:02:02.2')



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""
import pdb
import koppeltaal.interfaces
import koppeltaal.utils
import koppeltaal.models
import pytest
import requests
import selenium.webdriver.support.expected_conditions as EC
import selenium.webdriver.support.wait

from six.moves.urllib.parse import urlparse, parse_qs


def test_request_metadata(connector):
    result = connector.metadata()
    assert isinstance(result, dict)
    assert result.get('name') == 'Koppeltaal'
    assert result.get('version') >= 'v1.0'
    assert result.get('fhirVersion') == '0.0.82'


def wait_for_application(browser):
    # wait until the page redirect dance is over.
    selenium.webdriver.support.wait.WebDriverWait(
        browser, 10).until(lambda d: 'test.html' in d.current_url)


def login_with_oauth(browser):
    [b for b in browser.find_elements_by_tag_name('button') if
        b.text == 'log in with oauth'][0].click()
    selenium.webdriver.support.wait.WebDriverWait(
        browser, 10).until(
            EC.text_to_be_present_in_element(
                ('id', 'authorizationOutput'), 'access_token'))


def request_care_plan(browser):
    [b for b in browser.find_elements_by_tag_name('button') if
        b.text == 'request care plan'][0].click()
    selenium.webdriver.support.wait.WebDriverWait(
        browser, 10).until(
            EC.text_to_be_present_in_element(
                ('id', 'carePlanOutput'), 'reference'))


def parse_launch_url(url):
    parsed_url = urlparse(url)
    query = parse_qs(parsed_url.query)
    return query.get('iss', [''])[0]


def test_launch_patient(
        connector, careplan, careplan_response, patient, browser):
    launch_url = connector.launch(careplan, user=patient)
    assert parse_launch_url(launch_url).startswith(connector.transport.server)

    # There is a 'login with oauth' button in the page, let's see what that
    # does.
    browser.get(launch_url)
    wait_for_application(browser)
    selenium.webdriver.support.wait.WebDriverWait(
        browser, 10).until(
            lambda d: browser.find_element_by_id('patientReference'))

    assert browser.find_element_by_id('patientReference').text == ''
    assert browser.find_element_by_id('userReference').text == ''

    login_with_oauth(browser)
    selenium.webdriver.support.wait.WebDriverWait(
        browser, 10).until(
            lambda d: browser.find_element_by_id('patientReference'))
    assert browser.find_element_by_id('patientReference').text == \
        careplan.patient.fhir_link
    assert browser.find_element_by_id('userReference').text == \
        careplan.patient.fhir_link


def test_launch_practitioner(
        connector, careplan, careplan_response, practitioner, browser):
    # There is a 'login with oauth' button in the page, let's see what that
    # does.
    launch_url = connector.launch(careplan, user=practitioner)
    assert parse_launch_url(launch_url).startswith(connector.transport.server)

    browser.get(launch_url)
    wait_for_application(browser)
    assert browser.find_element_by_id('patientReference').text == ''
    assert browser.find_element_by_id('userReference').text == ''

    login_with_oauth(browser)
    assert browser.find_element_by_id('patientReference').text == \
        careplan.patient.fhir_link
    assert browser.find_element_by_id('userReference').text == \
        practitioner.fhir_link


def test_sso(connector):
    connector.integration.client_id = 'MindDistrict'
    connector.integration.client_secret = \
        connector._credentials.options.get('oauth_secret')

    patient_link = (
        'https://app.minddistrict.com/fhir/Koppeltaal/Patient/1394433515')
    step_1 = connector.launch_from_parameters(
        'MindDistrict', patient_link, patient_link, 'KTSTESTGAME')

    parts = urlparse(step_1)
    query = parse_qs(parts.query)

    assert query['application_id'][0] == 'MindDistrict'
    assert query['launch_id'][0] != ''

    step_2 = connector.authorize_from_parameters(
        query['application_id'][0],
        query['launch_id'][0],
        'https://example.com/koppeltaalauth')

    step_6 = requests.get(
        step_2, allow_redirects=False).headers.get('Location')

    parts = urlparse(step_6)
    query = parse_qs(parts.query)

    token = connector.token_from_parameters(
        query['code'][0],
        'https://example.com/koppeltaalauth')

    assert 'access_token' in token
    assert 'refresh_token' in token

    assert 'domain' in token and token['domain'] == connector.domain
    assert 'expires_in' in token and token['expires_in'] == 3600
    assert 'patient' in token and token['patient'] == patient_link
    assert 'resource' in token and token['resource'] == 'KTSTESTGAME'
    assert 'scope' in token and token['scope'] == 'patient/*.read'
    assert 'token_type' in token and token['token_type'] == 'Bearer'
    assert 'user' in token and token['user'] == patient_link


def test_send_activity(connector):
    uuid = koppeltaal.utils.uniqueid()

    assert len(list(connector.activities())) == 1
    assert connector.activity(u'uuid://{}'.format(uuid)) is None

    application = koppeltaal.models.ReferredResource(
        display='Test Generated Application Reference {}'.format(uuid))
    ad = koppeltaal.models.ActivityDefinition(
        application=application,
        description=u'Test Generated AD {}'.format(uuid),
        identifier=u'uuid://{}'.format(uuid),
        kind='ELearning',
        name=u'Test Generated AD {}'.format(uuid),
        performer='Patient',
        subactivities=[])

    updated = connector.send_activity(ad)
    assert len(list(connector.activities())) == 2
    assert updated.fhir_link is not None
    assert updated.fhir_link.startswith(
        'https://edgekoppeltaal.vhscloud.nl/FHIR/Koppeltaal'
        '/Other/ActivityDefinition:')
    assert updated.is_active is True
    assert updated.is_archived is False

    fetched = connector.activity(u'uuid://{}'.format(uuid))
    assert updated.fhir_link == fetched.fhir_link

    updated.is_active = False
    updated.is_archived = True

    connector.send_activity(updated)
    assert updated.fhir_link != fetched.fhir_link
    assert updated.fhir_link > fetched.fhir_link
    assert updated.is_active is False
    assert updated.is_archived is True
    assert connector.activity(u'uuid://{}'.format(uuid)) is None

    assert len(list(connector.activities())) == 1


def test_send_versioned_focal_resource_careplan(connector, careplan):
    response = connector.send(
        'CreateOrUpdateCarePlan', careplan, careplan.patient)
    first_version = response[0].fhir_link
    assert careplan.fhir_link != first_version
    assert careplan.fhir_link == \
        koppeltaal.utils.strip_history_from_link(first_version)

    with pytest.raises(koppeltaal.interfaces.OperationOutcomeError) as info:
        # Sending an update to the CarePlan without a version, now that the
        # server gave it a version, fails.
        connector.send(
            'CreateOrUpdateCarePlan', careplan, careplan.patient)

    assert info.value.outcome.issue[0].details == (
        'No version specified for the focal resource, message is rejected.')

    # When we properly update the CarePlan's version (FHIR link), we can send
    # an update again. And we get the subsequent version for it.
    careplan.fhir_link = first_version
    response = connector.send(
        'CreateOrUpdateCarePlan', careplan, careplan.patient)
    second_version = response[0].fhir_link
    assert first_version != second_version
    assert koppeltaal.utils.strip_history_from_link(first_version) == \
        koppeltaal.utils.strip_history_from_link(second_version)

    # Reusing the earlier version results in an error.
    careplan.fhir_link = first_version
    with pytest.raises(koppeltaal.interfaces.OperationOutcomeError) as info:
        connector.send(
            'CreateOrUpdateCarePlan', careplan, careplan.patient)

    assert info.value.outcome.issue[0].details == (
        'The specified resource version is not correct')

    careplan.fhir_link = second_version
    response = connector.send(
        'CreateOrUpdateCarePlan', careplan, careplan.patient)
    third_version = response[0].fhir_link
    assert first_version != second_version != third_version
    assert koppeltaal.utils.strip_history_from_link(first_version) == \
        koppeltaal.utils.strip_history_from_link(second_version) == \
        koppeltaal.utils.strip_history_from_link(third_version)


def test_send_versioned_focal_resource_patient(connector, patient):
    response = connector.send('CreateOrUpdatePatient', patient)
    first_version = response[0].fhir_link
    assert patient.fhir_link != first_version
    assert patient.fhir_link == \
        koppeltaal.utils.strip_history_from_link(first_version)

    with pytest.raises(koppeltaal.interfaces.OperationOutcomeError) as info:
        # Sending an update to the Patient without a version, now that the
        # server gave it a version, fails.
        connector.send('CreateOrUpdatePatient', patient)

    assert info.value.outcome.issue[0].details == (
        'No version specified for the focal resource, message is rejected.')

    # When we properly update the Patient's version (FHIR link), we can send
    # an update again. And we get the subsequent version for it.
    patient.fhir_link = first_version
    response = connector.send('CreateOrUpdatePatient', patient)
    second_version = response[0].fhir_link
    assert first_version != second_version
    assert koppeltaal.utils.strip_history_from_link(first_version) == \
        koppeltaal.utils.strip_history_from_link(second_version)

    # Reusing the earlier version results in an error.
    patient.fhir_link = first_version
    with pytest.raises(koppeltaal.interfaces.OperationOutcomeError) as info:
        response = connector.send('CreateOrUpdatePatient', patient)

    assert info.value.outcome.issue[0].details == (
        'The specified resource version is not correct')

    patient.fhir_link = second_version
    response = connector.send('CreateOrUpdatePatient', patient)
    third_version = response[0].fhir_link
    assert first_version != second_version != third_version
    assert koppeltaal.utils.strip_history_from_link(first_version) == \
        koppeltaal.utils.strip_history_from_link(second_version) == \
        koppeltaal.utils.strip_history_from_link(third_version)


def test_send_patient_resource_message(connector, patient):
    connector.send('CreateOrUpdatePatient', patient)


def test_send_practitioner_resource_message(connector, practitioner):
    connector.send('CreateOrUpdatePractitioner', practitioner)


def test_send_versioned_focal_resource_related_person(connector,
                                                      related_person):
    response = connector.send('CreateOrUpdateRelatedPerson',
        related_person, related_person.patient)
    first_version = response[0].fhir_link
    assert related_person.fhir_link != first_version
    assert related_person.fhir_link == \
        koppeltaal.utils.strip_history_from_link(first_version)

    with pytest.raises(koppeltaal.interfaces.OperationOutcomeError) as info:
        # Sending an update to the RelatedPerson without a version, now that
        # the server gave it a version, fails.
        connector.send('CreateOrUpdateRelatedPerson',
            related_person, related_person.patient)

    assert info.value.outcome.issue[0].details == (
        'No version specified for the focal resource, message is rejected.')

    # When we properly update the RelatedPerson's version (FHIR link), we
    # can send an update again. And we get the subsequent version for it.
    related_person.fhir_link = first_version
    response = connector.send('CreateOrUpdateRelatedPerson',
        related_person, related_person.patient)
    second_version = response[0].fhir_link
    assert first_version != second_version
    assert koppeltaal.utils.strip_history_from_link(first_version) == \
        koppeltaal.utils.strip_history_from_link(second_version)

    # Reusing the earlier version results in an error.
    related_person.fhir_link = first_version
    with pytest.raises(koppeltaal.interfaces.OperationOutcomeError) as info:
        response = connector.send('CreateOrUpdateRelatedPerson',
            related_person, related_person.patient)

    assert info.value.outcome.issue[0].details == (
        'The specified resource version is not correct')

    related_person.fhir_link = second_version
    response = connector.send('CreateOrUpdateRelatedPerson',
        related_person, related_person.patient)
    third_version = response[0].fhir_link
    assert first_version != second_version != third_version
    assert koppeltaal.utils.strip_history_from_link(first_version) == \
        koppeltaal.utils.strip_history_from_link(second_version) == \
        koppeltaal.utils.strip_history_from_link(third_version)


def test_send_related_person_resource_message(connector, related_person):
    connector.send('CreateOrUpdateRelatedPerson',
        related_person, related_person.patient)


def test_send_related_person_resource_in_careplan(
        connector, careplan, related_person):
    careplan.participants.append(
        koppeltaal.models.Participant(
            member=related_person,
            role='Thirdparty'))
    response = connector.send(
        'CreateOrUpdateCarePlan', careplan, careplan.patient)

    msghdr = [model for model in connector.search(patient=careplan.patient)][0]
    bundle = connector.search(message_id=msghdr.fhir_link)
    for m in bundle:
        if koppeltaal.definitions.RelatedPerson.providedBy(m):
            assert koppeltaal.utils.strip_history_from_link(m.fhir_link) == \
                koppeltaal.utils.strip_history_from_link(
                    related_person.fhir_link)



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import zope.interface

from koppeltaal import (interfaces, codes)


NULL_SYSTEM = 'http://hl7.org/fhir/v3/NullFlavor'
NULL_VALUE = 'UNK'

IDENTIFIER_AGB_Z = 'http://fhir.nl/fhir/NamingSystem/agb-z'
IDENTIFIER_BSN = 'http://fhir.nl/fhir/NamingSystem/bsn'

NAMING_SYSTEM_OFFICIAL = {
    IDENTIFIER_AGB_Z,
    IDENTIFIER_BSN,
}

FIELD_TYPES = {
    'boolean',
    'codeable',
    'code',
    'coding',
    'date',
    'datetime',
    'instant',
    'integer',
    'object',
    'reference',
    'string',
    'versioned reference',
}
RESERVED_NAMES = {
    'extension',
    'id',
    'language',
    'resourceType',
    'text',
}


class Field(zope.interface.Attribute):

    def __init__(
            self,
            name,
            field_type,
            binding=None,
            default=None,
            extension=None,
            multiple=False,
            optional=False,
            reserved_allowed=False):
        super(Field, self).__init__(__name__='')
        assert reserved_allowed or name not in RESERVED_NAMES, \
            '{} is a reserved name.'
        assert field_type in FIELD_TYPES, \
            'Unknown field type {} for {}.'.format(field_type, name)
        assert field_type not in {
            'object', 'codeable', 'code', 'coding'} or binding, \
            'Missing binding for {}.'.format(name)
        self.field_type = field_type
        self.name = name
        self.binding = binding
        self.default = default
        self.optional = optional
        self.multiple = multiple
        self.extension = extension
        self.url = None
        if extension:
            self.url = interfaces.NAMESPACE + extension

    def is_empty(self, value):
        if value is None:
            return True
        if self.multiple and isinstance(value, list) and len(value) == 0:
            return True
        return False


def extension_data_type(name):

    def data_type_iface(cls):
        cls.setTaggedValue('extension data type', 'value{}'.format(name))
        return cls

    return data_type_iface


def resource_type(name, standard=True):

    def resource_iface(cls):
        assert issubclass(cls, interfaces.IFHIRResource)
        cls.setTaggedValue('resource type', (name, standard))
        return cls

    return resource_iface


class SubActivityDefinition(zope.interface.Interface):

    name = Field(
        'name', 'string',
        extension='ActivityDefinition#SubActivityName')

    identifier = Field(
        'identifier', 'string',
        extension='ActivityDefinition#SubActivityIdentifier')

    description = Field(
        'description', 'string',
        optional=True,
        extension='ActivityDefinition#SubActivityDescription')

    active = Field(
        'isActive', 'boolean',
        default=True,
        extension='ActivityDefinition#SubActivityIsActive',
        optional=True)


@resource_type('ActivityDefinition', False)
class ActivityDefinition(interfaces.IIdentifiedFHIRResource):

    application = Field(
        'application', 'reference',
        extension='ActivityDefinition#Application')

    # Note that this is not required in the specification but his in
    # practice so that other messages can refer to it.
    identifier = Field(
        'activityDefinitionIdentifier', 'string',
        extension='ActivityDefinition#ActivityDefinitionIdentifier')

    kind = Field(
        'type', 'coding',
        binding=codes.ACTIVITY_KIND,
        extension='ActivityDefinition#ActivityKind')

    name = Field(
        'name', 'string',
        extension='ActivityDefinition#ActivityName')

    description = Field(
        'description', 'string',
        extension='ActivityDefinition#ActivityDescription',
        optional=True)

    subactivities = Field(
        'subActivity', 'object',
        binding=SubActivityDefinition,
        extension='ActivityDefinition#SubActivity',
        multiple=True,
        optional=True)

    performer = Field(
        'defaultPerformer', 'coding',
        binding=codes.ACTIVITY_PERFORMER,
        extension='ActivityDefinition#DefaultPerformer',
        optional=True)

    launch_type = Field(
        'launchType', 'code',
        binding=codes.ACTIVITY_LAUNCH_TYPE,
        default='Web',
        extension='ActivityDefinition#LaunchType',
        optional=True)

    is_active = Field(
        'isActive', 'boolean',
        default=True,
        extension='ActivityDefinition#IsActive',
        optional=True)

    is_domain_specific = Field(
        'isDomainSpecific', 'boolean',
        extension='ActivityDefinition#IsDomainSpecific',
        optional=True)

    is_archived = Field(
        'isArchived', 'boolean',
        default=False,
        extension='ActivityDefinition#IsArchived',
        optional=True)


class CarePlanSubActivityStatus(zope.interface.Interface):

    definition = Field(
        'identifier', 'string',
        extension='CarePlanActivityStatus#SubActivityIdentifier')

    status = Field(
        'status', 'coding',
        binding=codes.CAREPLAN_ACTIVITY_STATUS,
        extension='CarePlanActivityStatus#SubActivityStatus')


@resource_type('CarePlanActivityStatus', False)
class CarePlanActivityStatus(interfaces.IIdentifiedFHIRResource):

    identifier = Field(
        'activity', 'string',
        extension='CarePlanActivityStatus#Activity')

    status = Field(
        'activityStatus', 'coding',
        binding=codes.CAREPLAN_ACTIVITY_STATUS,
        extension='CarePlanActivityStatus#ActivityStatus')

    subactivities = Field(
        'subactivity', 'object',
        binding=CarePlanSubActivityStatus,
        extension='CarePlanActivityStatus#SubActivity',
        multiple=True,
        optional=True)

    percentage = Field(
        'percentageCompleted', 'integer',
        extension='CarePlanActivityStatus#PercentageCompleted',
        optional=True)


@extension_data_type('HumanName')
class Name(zope.interface.Interface):

    # http://hl7.org/fhir/DSTU1/datatypes.html#HumanName

    given = Field(
        'given', 'string',
        multiple=True,
        optional=True)

    family = Field(
        'family', 'string',
        multiple=True,
        optional=True)

    prefix = Field(
        'prefix', 'string',
        multiple=True,
        optional=True)

    suffix = Field(
        'suffix', 'string',
        multiple=True,
        optional=True)

    use = Field(
        'use', 'code',
        binding=codes.NAME_USE,
        optional=True)

    text = Field(
        'text', 'string',
        optional=True,
        reserved_allowed=True)


@extension_data_type('Contact')
class Contact(zope.interface.Interface):

    system = Field(
        'system', 'code',
        binding=codes.CONTACT_SYSTEM,
        optional=True)

    value = Field(
        'value', 'string',
        optional=True)

    use = Field(
        'use', 'code',
        binding=codes.CONTACT_USE,
        optional=True)


@extension_data_type('Identifier')
class Identifier(zope.interface.Interface):

    system = Field(
        'system', 'string',
        optional=True)

    value = Field(
        'value', 'string',
        optional=True)

    use = Field(
        'use', 'code',
        binding=codes.IDENTIFIER_USE,
        optional=True)


class OrganizationContactPerson(zope.interface.Interface):

    contacts = Field(
        'telecom', 'object',
        binding=Contact,
        multiple=True,
        optional=True)

    gender = Field(
        'gender', 'codeable',
        binding=codes.GENDER,
        optional=True)

    name = Field(
        'name', 'object',
        binding=Name,
        optional=True)

    purpose = Field(
        'purpose', 'code',
        binding=codes.CONTACT_ENTITY_TYPE,
        optional=True)


@extension_data_type('Period')
class Period(zope.interface.Interface):

    start = Field(
        'start', 'datetime',
        optional=True)

    end = Field(
        'end', 'datetime',
        optional=True)


@extension_data_type('Address')
class Address(zope.interface.Interface):

    city = Field(
        'city', 'string',
        optional=True)

    country = Field(
        'country', 'string',
        optional=True)

    line = Field(
        'line', 'string',
        multiple=True,
        optional=True)

    period = Field(
        'period', 'object',
        binding=Period,
        optional=True)

    state = Field(
        'state', 'string',
        optional=True)

    text = Field(
        'text', 'string',
        optional=True,
        reserved_allowed=True)

    use = Field(
        'use', 'code',
        binding=codes.ADDRESS_USE,
        optional=True)

    zip = Field(
        'zip', 'string',
        optional=True)


@resource_type('Organization')
class Organization(interfaces.IIdentifiedFHIRResource):

    active = Field(
        'active', 'boolean',
        optional=True)

    address = Field(
        'address', 'object',
        binding=Address,
        multiple=True,
        optional=True)

    category = Field(
        'type', 'code',
        binding=codes.ORGANIZATION_TYPE,
        optional=True)

    contacts = Field(
        'telecom', 'object',
        binding=Contact,
        multiple=True,
        optional=True)

    contact_persons = Field(
        'contact', 'object',
        binding=OrganizationContactPerson,
        multiple=True,
        optional=True)

    identifiers = Field(
        'identifier', 'object',
        binding=Identifier,
        multiple=True,
        optional=True)

    name = Field(
        'name', 'string',
        optional=True)

    part_of = Field(
        'partOf', 'reference',
        optional=True)


@resource_type('Patient')
class Patient(interfaces.IIdentifiedFHIRResource):

    active = Field(
        'active', 'boolean',
        optional=True)

    address = Field(
        'address', 'object',
        binding=Address,
        multiple=True,
        optional=True)

    age = Field(
        'age', 'integer',
        optional=True,
        extension='Patient#Age')

    birth_date = Field(
        'birthDate', 'datetime',
        optional=True)

    care_providers = Field(
        'careProvider', 'reference',
        multiple=True,
        optional=True)

    contacts = Field(
        'telecom', 'object',
        binding=Contact,
        multiple=True,
        optional=True)

    identifiers = Field(
        'identifier', 'object',
        binding=Identifier,
        multiple=True,
        optional=True)

    gender = Field(
        'gender', 'codeable',
        binding=codes.GENDER,
        optional=True)

    name = Field(
        'name', 'object',
        binding=Name,
        multiple=True)

    managing_organization = Field(
        'managingOrganization', 'reference',
        optional=True)


@resource_type('Practitioner')
class Practitioner(interfaces.IIdentifiedFHIRResource):

    birth_date = Field(
        'birthDate', 'datetime',
        optional=True)

    contacts = Field(
        'telecom', 'object',
        binding=Contact,
        multiple=True,
        optional=True)

    identifiers = Field(
        'identifier', 'object',
        binding=Identifier,
        multiple=True,
        optional=True)

    name = Field(
        'name', 'object',
        binding=Name,
        optional=True)

    gender = Field(
        'gender', 'codeable',
        binding=codes.GENDER,
        optional=True)

    organization = Field(
        'organization', 'reference',
        optional=True)


@resource_type('RelatedPerson')
class RelatedPerson(interfaces.IIdentifiedFHIRResource):

    identifiers = Field(
        'identifier', 'object',
        binding=Identifier,
        multiple=True,
        optional=True)

    patient = Field(
        'patient', 'reference')

    relationship = Field(
        'relationship', 'codeable',
        binding=codes.RELATED_PERSON_RELATIONSHIP_TYPE,
        optional=True)

    name = Field(
        'name', 'object',
        binding=Name,
        optional=True)

    contacts = Field(
        'telecom', 'object',
        binding=Contact,
        multiple=True,
        optional=True)

    gender = Field(
        'gender', 'codeable',
        binding=codes.GENDER,
        optional=True)

    address = Field(
        'address', 'object',
        binding=Address,
        optional=True)


class Participant(zope.interface.Interface):

    member = Field(
        'member', 'reference')

    role = Field(
        'role', 'codeable',
        binding=codes.CAREPLAN_PARTICIPANT_ROLE,
        optional=True)

    careteam = Field(
        'careteam', 'reference',
        extension='CarePlan#ParticipantCareTeam',
        multiple=True,
        optional=True)


class Goal(zope.interface.Interface):

    description = Field(
        'description', 'string')

    status = Field(
        'status', 'code',
        binding=codes.CAREPLAN_GOAL_STATUS,
        optional=True)

    notes = Field(
        'notes', 'string',
        optional=True)


class SubActivity(zope.interface.Interface):

    # Note how this definition "points" to the `identifier` one of the
    # `ActivityDefinition.subActivity`.
    definition = Field(
        'identifier', 'string',
        extension='CarePlan#SubActivityIdentifier')

    status = Field(
        'status', 'coding',
        binding=codes.CAREPLAN_ACTIVITY_STATUS,
        extension='CarePlan#SubActivityStatus',
        optional=True)


class ActivityParticipant(zope.interface.Interface):

    member = Field(
        'member', 'reference',
        extension='CarePlan#ParticipantMember')

    role = Field(
        'role', 'codeable',
        binding=codes.CAREPLAN_PARTICIPANT_ROLE,
        extension='CarePlan#ParticipantRole',
        optional=True)

    careteam = Field(
        'careteam', 'reference',
        extension='CarePlan#ParticipantCareTeam',
        multiple=True,
        optional=True)


class ActivityDetails(zope.interface.Interface):

    category = Field(
        'category', 'code',
        binding=codes.CAREPLAN_ACTIVITY_CATEGORY)

    performers = Field(
        'performer', 'reference',
        multiple=True,
        optional=True)


class Activity(zope.interface.Interface):

    # This is optional in the specification but cannot be in practice,
    # or other messages won't be able to refer to it.
    identifier = Field(
        'identifier', 'string',
        extension='CarePlan#ActivityIdentifier')

    cancelled = Field(
        'cancelled', 'instant',
        extension='CarePlan#Cancelled',
        optional=True)

    # Note how this definition "points" to the `identifier` one of the
    # `ActivityDefinition`.
    definition = Field(
        'definition', 'string',
        extension='CarePlan#ActivityDefinition',
        optional=True)

    description = Field(
        'description', 'string',
        extension='CarePlan#ActivityDescription',
        optional=True)

    ends = Field(
        'endDate', 'datetime',
        extension='CarePlan#EndDate',
        optional=True)

    finished = Field(
        'finished', 'instant',
        extension='CarePlan#Finished',
        optional=True)

    # Note the `kind` should match the `kind` of the `ActivityDefinition`
    # we're pointing to.
    kind = Field(
        'type', 'coding',
        binding=codes.ACTIVITY_KIND,
        extension='CarePlan#ActivityKind')

    notes = Field(
        'notes', 'string',
        optional=True)

    participants = Field(
        'participant', 'object',
        binding=ActivityParticipant,
        extension='CarePlan#Participant',
        multiple=True,
        optional=True)

    planned = Field(
        'startDate', 'datetime',
        extension='CarePlan#StartDate')

    started = Field(
        'started', 'instant',
        optional=True,
        extension='CarePlan#Started')

    # This is the older version of the `status`. KT 1.1.1 uses a `code` and
    # points to the http://hl7.org/fhir/care-plan-activity-status value set.
    # Perhaps we can update it and still be compatible with Kickass Game and
    # more importantly, with KT 1.1.1.
    status = Field(
        'status', 'coding',
        binding=codes.CAREPLAN_ACTIVITY_STATUS,
        extension='CarePlan#ActivityStatus')

    subactivities = Field(
        'subactivity', 'object',
        binding=SubActivity,
        extension='CarePlan#SubActivity',
        multiple=True,
        optional=True)

    details = Field(
        'simple', 'object',
        binding=ActivityDetails,
        optional=True)

    prohibited = Field(
        'prohibited', 'boolean',
        default=False,
        optional=True)


@resource_type('CarePlan')
class CarePlan(interfaces.IIdentifiedFHIRResource):

    activities = Field(
        'activity', 'object',
        binding=Activity,
        multiple=True,
        optional=True)

    goals = Field(
        'goal', 'object',
        binding=Goal,
        multiple=True,
        optional=True)

    participants = Field(
        'participant', 'object',
        binding=Participant,
        multiple=True,
        optional=True)

    patient = Field(
        'patient', 'reference',
        optional=True)

    status = Field(
        'status', 'code',
        binding=codes.CAREPLAN_STATUS)


class ProcessingStatus(zope.interface.Interface):

    status = Field(
        'status', 'code',
        binding=codes.PROCESSING_STATUS,
        extension='MessageHeader#ProcessingStatusStatus',
        optional=True)

    last_changed = Field(
        'statusLastChanged', 'instant',
        extension='MessageHeader#ProcessingStatusStatusLastChanged')

    exception = Field(
        'exception', 'string',
        extension='MessageHeader#ProcessingStatusException',
        optional=True)


class MessageHeaderSource(zope.interface.Interface):

    name = Field(
        'name', 'string',
        optional=True)

    software = Field('software', 'string')

    version = Field(
        'version', 'string',
        optional=True)

    endpoint = Field('endpoint', 'string')


class MessageHeaderResponse(zope.interface.Interface):

    identifier = Field('identifier', 'string')

    code = Field(
        'code', 'code',
        binding=codes.MESSAGE_HEADER_RESPONSE_CODE)


@resource_type('MessageHeader')
class MessageHeader(interfaces.IFHIRResource):

    identifier = Field(
        'identifier', 'string')

    timestamp = Field(
        'timestamp', 'instant')

    data = Field(
        'data', 'versioned reference',
        multiple=True,
        optional=True)

    patient = Field(
        'patient', 'reference',
        extension='MessageHeader#Patient',
        optional=True)

    event = Field(
        'event', 'coding',
        binding=codes.MESSAGE_HEADER_EVENTS)

    response = Field(
        'response', 'object',
        binding=MessageHeaderResponse,
        optional=True)

    status = Field(
        'processingStatus', 'object',
        binding=ProcessingStatus,
        extension='MessageHeader#ProcessingStatus',
        optional=True)

    source = Field(
        'source', 'object',
        binding=MessageHeaderSource,
        optional=True)


@resource_type('CareTeam', standard=False)
class CareTeam(interfaces.IIdentifiedFHIRResource):

    identifier = Field(
        'careTeamIdentifier', 'object',
        binding=Identifier,
        extension='CareTeam#CareTeamIdentifier',
        multiple=True,
        optional=True)

    status = Field(
        'status', 'coding',
        binding=codes.CARE_TEAM_STATUS,
        extension='CareTeam#Status',
        optional=True)

    name = Field(
        'name', 'string',
        extension='CareTeam#Name',
        optional=True)

    subject = Field(
        'subject', 'reference',
        extension='CareTeam#Subject',
        optional=True)

    period = Field(
        'period', 'object',
        binding=Period,
        extension='CareTeam#Period',
        optional=True)

    managing_organization = Field(
        'managingOrganization', 'reference',
        extension='CareTeam#ManagingOrganization',
        multiple=True,
        optional=True)


class Issue(zope.interface.Interface):

    severity = Field(
        'severity', 'code',
        binding=codes.ISSUE_SEVERITY)

    type = Field(
        'type', 'coding',
        binding=codes.ISSUE_TYPE,
        optional=True)

    resource = Field(
        'resource', 'reference',
        extension='OperationOutcome#IssueResource',
        optional=True)

    details = Field(
        'details', 'string',
        optional=True)

    location = Field(
        'location', 'string',
        optional=True,
        multiple=True)


@resource_type('OperationOutcome')
class OperationOutcome(interfaces.IIdentifiedFHIRResource):

    issue = Field(
        'issue', 'object',
        binding=Issue,
        multiple=True)
